@@ -1,19 +1,50 @@
 package CPAN::Reporter;
 use strict;
 
-$CPAN::Reporter::VERSION = $CPAN::Reporter::VERSION = "0.28";
+$CPAN::Reporter::VERSION = $CPAN::Reporter::VERSION = "0.29";
 
 use Config;
 use Config::Tiny ();
 use ExtUtils::MakeMaker qw/prompt/;
 use File::Basename qw/basename/;
 use File::HomeDir ();
-use File::Path qw/mkpath/;
+use File::Path qw/mkpath rmtree/;
 use File::Temp ();
 use Probe::Perl ();
 use Tee qw/tee/;
 use Test::Reporter ();
 
+#--------------------------------------------------------------------------#
+# Back-compatibility checks -- just once per load
+#--------------------------------------------------------------------------#
+
+# 0.28_51 changed Mac OS X config file location -- if old directory is found,
+# move it to the new location
+if ( $^O eq 'darwin' ) {
+    my $old = File::Spec->catdir(File::HomeDir->my_documents,".cpanreporter");
+    my $new = File::Spec->catdir(File::HomeDir->my_home,".cpanreporter");
+    if ( ( -d $old ) && (! -d $new ) ) {
+        print STDERR << "HERE";
+Since CPAN::Reporter 0.28_51, the Mac OSX config directory has changed. 
+
+  Old: $old
+  New: $new  
+
+Your existing configuration file will be moved automatically.
+HERE
+        mkpath($new);
+        open OLD_CONFIG, File::Spec->catfile($old, "config.ini")
+            or die $!;
+        open NEW_CONFIG, ">" . File::Spec->catfile($new, "config.ini")
+            or die $!;
+        print NEW_CONFIG <OLD_CONFIG>;
+        close OLD_CONFIG;
+        close NEW_CONFIG;
+        unlink File::Spec->catfile($old, "config.ini") or die $!;
+        rmdir($old) or die $!;
+    }
+}
+
 #--------------------------------------------------------------------------#
 # defaults and prompts
 #--------------------------------------------------------------------------#
@@ -22,6 +53,24 @@ use Test::Reporter ();
 
 my @config_order = qw/ email_from cc_author edit_report send_report
                        smtp_server /;
+
+my $grade_action_prompt = << 'HERE'; 
+
+Some of the following configuration options require one or more "grade:action"
+pairs that determine what grade-specific action to take for that option.
+These pairs should be space-separated and are processed left-to-right. See
+CPAN::Reporter documentation for more details.
+
+    GRADE   :   ACTION  ======> EXAMPLES        
+    -------     -------         --------    
+    pass        yes             default:no
+    fail        no              default:yes pass:no
+    unknown     ask/no          default:ask/no pass:yes fail:no
+    na          ask/yes         
+    default
+
+HERE
+
 my %defaults = (
     email_from => {
         default => '',
@@ -34,34 +83,37 @@ by your email address in angle marks, e.g. "John Doe" <jdoe@nowhere.com>
 HERE
     },
     cc_author => {
-        default => 'fail',
+        default => 'default:yes pass:no',
         prompt => "Do you want to CC the the module author?",
+        validate => 1,
         info => <<'HERE',
 If you would like, CPAN::Reporter will copy the module author with
 the results of your tests.  By default, authors are copied only on 
-failed/unknown results. This option takes a "yes/no/fail/ask" value.  
+failed/unknown results. This option takes "grade:action" pairs.  
 HERE
     },
     edit_report => {
-        default => 'ask/no',
+        default => 'default:ask/no pass:no',
         prompt => "Do you want to edit the test report?",
+        validate => 1,
         info => <<'HERE',
 Before test reports are sent, you may want to edit the test report
 and add additional comments about the result or about your system or
 Perl configuration.  By default, CPAN::Reporter will ask after
 each report is generated whether or not you would like to edit the 
-report. This option takes a "yes/no/fail/ask" value.
+report. This option takes "grade:action" pairs.
 HERE
     },
     send_report => {
-        default => 'ask/yes',
+        default => 'default:ask/yes pass:yes',
         prompt => "Do you want to send the test report?",
+        validate => 1,
         info => <<'HERE',
 By default, CPAN::Reporter will prompt you for confirmation that
 the test report should be sent before actually emailing the 
 report.  This gives the opportunity to bypass sending particular
 reports if you need to (e.g. a duplicate of an earlier result).
-This option takes a "yes/no/fail/ask" value.
+This option takes "grade:action" pairs.
 HERE
     },
     smtp_server => {
@@ -100,9 +152,12 @@ sub configure {
     my $config;
     my $existing_options;
     
+    # explain grade:action pairs
+    print $grade_action_prompt;
+    
     # read or create
     if ( -f $config_file ) {
-        print "\nFound your CPAN::Reporter config file at '$config_file'.\n";
+        print "\nFound your CPAN::Reporter config file at:\n$config_file\n";
         $config = _open_config_file() 
             or return;
         $existing_options = _get_config_options( $config );
@@ -113,20 +168,34 @@ sub configure {
         $config = Config::Tiny->new();
     }
     
-    # initialize options that have an info description
     for my $k ( @config_order ) {
         my $option_data = $defaults{$k};
         print "\n" . $option_data->{info}. "\n";
+        # options with defaults are mandatory
         if ( defined $defaults{$k}{default} ) {
-            $config->{_}{$k} = prompt( 
+            # repeat until validated
+            PROMPT:
+            while ( my $answer = prompt(
                 "$k?", 
-                $existing_options->{$k} || $option_data->{default} 
-            );
+                $existing_options->{$k} || $option_data->{default} )
+            ) 
+            {
+                if ( $defaults{$k}{validate} ) {
+                    for my $ga ( split q{ }, $answer ) {
+                        if ( ! _validate_grade_action( $ga ) ) {
+                            warn "\nInvalid option '$ga' in '$k'\n\n";
+                            next PROMPT;
+                        }
+                    }
+                }
+                $config->{_}{$k} = $answer;
+                last PROMPT;
+            }
         }
         else {
-            # only initialize options with undef default if
-            # answer matches non white space, otherwise
-            # reset it
+            # only initialize options without default if
+            # answer matches non white space and validates, 
+            # otherwise reset it
             my $answer = prompt( 
                 "$k?", 
                 $existing_options->{$k} || q{} 
@@ -190,7 +259,7 @@ sub test {
         warn "CPAN::Reporter couldn't read test results\n";
         return;
     }
-    $result->{output} = do { local $/; <TEST_RESULT> };
+    $result->{output} = [ <TEST_RESULT> ];
     close TEST_RESULT;
 
     _process_report( $result );
@@ -201,16 +270,26 @@ sub test {
 # private functions
 #--------------------------------------------------------------------------#
 
+#--------------------------------------------------------------------------#
+# _get_config_dir
+#--------------------------------------------------------------------------#
+
 sub _get_config_dir {
-    return File::Spec->catdir(File::HomeDir->my_documents, ".cpanreporter");
+    return ( $^O eq 'MSWin32' )
+        ? File::Spec->catdir(File::HomeDir->my_documents, ".cpanreporter")
+        : File::Spec->catdir(File::HomeDir->my_home, ".cpanreporter") ;
 }
 
+#--------------------------------------------------------------------------#
+# _get_config_file
 #--------------------------------------------------------------------------#
 
 sub _get_config_file {
     return File::Spec->catdir( _get_config_dir, "config.ini" );
 }
 
+#--------------------------------------------------------------------------#
+# _get_config_options
 #--------------------------------------------------------------------------#
 
 sub _get_config_options {
@@ -229,6 +308,8 @@ sub _get_config_options {
     return \%active;
 }
 
+#--------------------------------------------------------------------------#
+# _grade_msg
 #--------------------------------------------------------------------------#
 
 sub _grade_msg {
@@ -239,46 +320,61 @@ sub _grade_msg {
     return;
 }
 
+#--------------------------------------------------------------------------#
+# _grade_report
 #--------------------------------------------------------------------------#
 
 sub _grade_report {
     my $result = shift;
     my ($grade,$is_make,$msg);
-
-    # CPAN.pm won't normally test a failed 'make', so that should
-    # catch prereq failures that would normally be "unknown".
-    # XXX really should check in case test was forced
-    #
+    my $output = $result->{output};
+    
     # Output strings taken from Test::Harness::
     # _show_results()  -- for versions < 2.57_03 
     # get_results()    -- for versions >= 2.57_03
 
+    # XXX don't shortcut to unknown with _has_tests here because a custom
+    # Makefile.PL or Build.PL might define tests in a non-standard way
+    
     # check for make or Build
     $is_make = _is_make( $result->{command} );
-        
-    # parse for Test::Harness results
-    if ( $result->{output} =~ m{^All tests successful}ms ) {
-        $grade = 'pass';
-        $msg = 'All tests successful';
+    
+    # parse in reverse order for Test::Harness results
+    for my $i ( reverse 0 .. $#{$output} ) {
+        if ( $output->[$i] =~ m{^All tests successful}ms ) {
+            $grade = 'pass';
+            $msg = 'All tests successful';
+        }
+        elsif ( $output->[$i] =~ m{^.?No tests defined}ms ) {
+            $grade = 'unknown';
+            $msg = 'No tests provided';
+        }
+        elsif ( $output->[$i] =~ m{^FAILED--no tests were run}ms ) {
+            $grade = 'unknown';
+            $msg = 'No tests were run';
+        }
+        elsif ( $output->[$i] =~ m{^FAILED--.*--no output}ms ) {
+            $grade = 'fail';
+            $msg = 'Tests had no output';
+        }
+        elsif ( $output->[$i] =~ m{^Failed }ms ) {  # must be lowercase
+            $grade = 'fail';
+            $msg = "Distribution had failing tests";
+        }
+        else {
+            next;
+        }
+        if ( $grade eq 'unknown' && _has_tests() ) {
+            # probably a spurious message from recursive make, so ignore and
+            # continue if we can find any standard test files
+            $grade = $msg = undef;
+            next;
+        }
+        last if $grade;
     }
-    elsif ( $result->{output} =~ m{^.?No tests defined}ms ) {
-        $grade = 'unknown';
-        $msg = 'No tests provided';
-    }
-    elsif ( $result->{output} =~ m{^FAILED--no tests were run}ms ) {
-        $grade = 'unknown';
-        $msg = 'No tests were run';
-    }
-    elsif ( $result->{output} =~ m{^FAILED--.*--no output}ms ) {
-        $grade = 'fail';
-        $msg = 'Tests had no output';
-    }
-    elsif ( $result->{output} =~ m{^Failed }ms ) {  # must be lowercase
-        $grade = 'fail';
-        $msg = "Distribution had failing tests";
-    }
-    else {
-        # didn't find Test::Harness output we recognized
+    
+    # didn't find Test::Harness output we recognized
+    if ( ! $grade ) {
         $grade = "unknown";
         $msg = "Couldn't determine a result";
     }
@@ -288,7 +384,7 @@ sub _grade_report {
     # so re-run make test on test.pl
     
     if ( $is_make && -f "test.pl" && $grade ne 'fail' ) {
-        if ( $result->{output} =~ m{^makewrapper: make failed}ims ) {
+        if ( $output->[$#{$output}] =~ m{^makewrapper: make failed}ims ) {
             $grade = "fail";
             $msg = "'make test' error detected";
         }
@@ -303,14 +399,11 @@ sub _grade_report {
 
     if ( $grade eq 'fail' ) {
         # check for perl version prerequisite or outright failure
-        if (
-            $result->{prereq_pm} =~ m{^\s+!\s+perl\s}ims ||
-            $result->{output} =~ m{Perl .*? required.*?this is only}ms
-        ) {
+        if ( $result->{prereq_pm} =~ m{^\s+!\s+perl\s}ims ) {
             $grade = 'na';
             $msg = 'Perl version too low';
         }
-        # check the prereq report for a failure flag (!)
+        # check the prereq report for missing or failure flag '!'
         elsif ( $result->{prereq_pm} =~ m{n/a}ims ) {
             $grade = 'na';
             $msg = 'Prerequisite missing';
@@ -325,6 +418,27 @@ sub _grade_report {
     return $grade;
 }
 
+#--------------------------------------------------------------------------#
+# _has_tests
+#--------------------------------------------------------------------------#
+
+sub _has_tests {
+    return 1 if -f 'test.pl';
+    if ( -d 't' ) {
+        local *TESTDIR;
+        opendir TESTDIR, 't';
+        while ( my $f = readdir TESTDIR ) {
+            if ( $f =~ m{\.t$} ) {
+                close TESTDIR;
+                return 1;
+            }
+        }
+    }
+    return 0;
+}
+
+#--------------------------------------------------------------------------#
+# _is_make
 #--------------------------------------------------------------------------#
 
 sub _is_make {
@@ -332,6 +446,29 @@ sub _is_make {
     return 1 if $command =~ m{^\S*make}ims;
 }
 
+#--------------------------------------------------------------------------#
+# _is_valid_action
+#--------------------------------------------------------------------------#
+
+my @valid_actions = qw{ yes no ask/yes ask/no ask };
+
+sub _is_valid_action {
+    my $action = shift;
+    return grep { $action eq $_ } @valid_actions;
+}
+
+#--------------------------------------------------------------------------#
+# _is_valid_grade
+#--------------------------------------------------------------------------#
+
+my @valid_grades = qw{ pass fail unknown na default };
+sub _is_valid_grade {
+    my $grade = shift;
+    return grep { $grade eq $_ } @valid_grades;
+}
+
+#--------------------------------------------------------------------------#
+# _open_config_file
 #--------------------------------------------------------------------------#
 
 sub _open_config_file {
@@ -342,6 +479,34 @@ sub _open_config_file {
     return $config; 
 }
 
+#--------------------------------------------------------------------------#
+# _parse_option
+#--------------------------------------------------------------------------#
+
+sub _parse_option {
+    my $name = shift;
+    my $input_string = "default:no " . shift;
+    
+    # preset defaults
+    my @options;
+
+    # process space-separated terms in order
+    for my $spec ( split q{ }, $input_string ) {
+        my ($grade_list,$action);
+        
+        # get valid parts or warn
+        my @grade_actions = _validate_grade_action($spec);
+        
+        if( ! @grade_actions ) {
+            warn "Ignoring invalid grade:action '$spec' for '$name'\n";
+        }
+        
+        push @options, @grade_actions;
+    }
+    
+    return { @options };
+}
+
 #--------------------------------------------------------------------------#
 # _prereq_report
 #--------------------------------------------------------------------------#
@@ -451,6 +616,8 @@ sub _prereq_report {
     return $report || "    No requirements found\n";
 }
 
+#--------------------------------------------------------------------------#
+# _process_report
 #--------------------------------------------------------------------------#
 
 sub _process_report {
@@ -537,27 +704,30 @@ EMAIL_REQUIRED
 
 sub _prompt {
     my ($config, $option, $grade) = @_;
+
+    my $dispatch = _parse_option( $option, $config->{$option} );
+    my $action = $dispatch->{$grade} || $dispatch->{default};
+
     my $prompt;
-    if     ( lc $config->{$option} eq 'ask/yes' ) { 
+    if     ( $action =~ m{^ask/yes}i ) { 
         $prompt = prompt( $defaults{$option}{prompt} . " (yes/no)", "yes" );
     }
-    elsif  ( $config->{$option} =~ m{^ask(/no)?}i ) {
+    elsif  ( $action =~ m{^ask(/no)?}i ) {
         $prompt = prompt( $defaults{$option}{prompt} . " (yes/no)", "no" );
     }
-    elsif  ( lc $config->{$option} =~ 'fail' ) {
-        $prompt = ( $grade =~ m{^(fail|unknown)$}i ) ? 'yes' : 'no';
-    }
     else { 
-        $prompt = $config->{$option};
+        $prompt = $action;
     }
     return lc $prompt;
 }
 
+#--------------------------------------------------------------------------#
+# _report_text
 #--------------------------------------------------------------------------#
 
 sub _report_text {
     my $data = shift;
-    
+    my $test_log = join(q{},@{$data->{output}});
     # generate report
     my $output = << "ENDREPORT";
 Dear $data->{author},
@@ -607,12 +777,59 @@ TEST OUTPUT
 
 Output from '$data->{command}':
 
-$data->{output}
+$test_log
 ENDREPORT
 
     return $output;
 }
 
+#--------------------------------------------------------------------------#
+# _validate_grade_action 
+# returns grade, action, grade, action ...
+# returns empty list/undef if invalid
+#--------------------------------------------------------------------------#
+
+sub _validate_grade_action {
+    my $grade_action = shift;
+    
+    my ($grade_list,$action);
+
+    if ( $grade_action =~ m{.:.} ) {
+        # parse pair for later check
+        ($grade_list, $action) = $grade_action =~ m{\A([^:]+):(.+)\z};
+    }
+    elsif ( _is_valid_action($grade_action) ) {
+        # action by itself
+        return "default", $grade_action;
+    }
+    elsif ( _is_valid_grade($grade_action) ) {
+        # grade by itself
+        return $grade_action, "yes";
+    }
+    elsif( $grade_action =~ m{./.} ) {
+        # gradelist by itself, so setup for later check
+        $grade_list = $grade_action;
+        $action = "yes";
+    }
+    else {
+        # something weird, so fail
+        return;
+    }
+        
+    # check gradelist
+    my @grades = split "/", $grade_list;
+    
+    for my $g ( @grades ) { 
+        return if ! _is_valid_grade($g);
+    }
+    
+    # check action
+    return if ! _is_valid_action($action);
+
+    # otherwise, it all must be OK
+    return map { $_ => $action } @grades;
+}
+
 1; #this line is important and will help the module return a true value
 
 __END__
@@ -697,6 +914,9 @@ died because of missing prerequisites
 * {unknown} -- no test files could be found (either t/*.t or test.pl) or 
 a result could not be determined from test output
 
+* {default} -- this is not an actual grade reported to CPAN Testers, but it
+is used in action prompt configuration options to indicate a fallback action
+
 In returning results to CPAN.pm, "pass" and "unknown" are considered successful
 attempts to "make test" or "Build test" and will not prevent installation.
 "fail" and "na" are considered to be failures and CPAN.pm will not install
@@ -705,10 +925,8 @@ unless forced.
 = CONFIG FILE OPTIONS
 
 Default options for CPAN::Reporter are read from a configuration file 
-{.cpanreporter/config.ini} in the user's home directory (Unix), "My 
-Documents" directory (Windows) or "~/Documents" directory (OS X).  
-(See [File::HomeDir] and the {my_documents} method for config folder
-location on other operating systems.)
+{.cpanreporter/config.ini} in the user's home directory (Unix and OS X)
+or "My Documents" directory (Windows).
 
 The configuration file is in "ini" format, with the option name and value
 separated by an "=" sign
@@ -716,26 +934,15 @@ separated by an "=" sign
   email_from = "John Doe" <johndoe@nowhere.org>
   cc_author = no
 
-Options shown below as taking "yes/no/fail/ask" should be set to one of
-five values; the result of each is as follows:
-
-* {yes} -- automatic yes
-* {no} -- automatic no
-* {fail} -- yes if the test result was failure/unknown; no otherwise
-* {ask/no} or just {ask} -- prompt each time, but default to no
-* {ask/yes} -- prompt each time, but default to yes
-
-For prompts, the default will be used if return is pressed immediately at
-the prompt or if the {PERL_MM_USE_DEFAULT} environment variable is set to
-a true value.
-
-Interactive configuration of required and standard options may be repeated at
-any time from the CPAN shell.  Interactive configuration will also includes 
-any additional options that already exist within the configuration file.
+Interactive configuration of email address, action prompts and mail server
+options may be repeated at any time from the CPAN shell.  
 
  cpan> o conf init test_report
 
-Descriptions for each option follow.
+Interactive configuration will also include any additional, non-standard
+options that have been added manually to the configuration file.
+
+Available options are described in the following sections.
 
 == Email Address (required)
 
@@ -757,42 +964,93 @@ Subscribing an account to the cpan-testers list is as easy as sending a blank
 email to cpan-testers-subscribe@perl.org and replying to the confirmation
 email.
 
-== Standard Options
+== Action Prompts
 
-These options are included in the standard config file template that is
-automatically created.
+Several steps in the generation of a test report are optional.  Configuration
+options control whether an action should be taken automatically or whether
+CPAN::Reporter should prompt the user for the action to take.  The action
+to take may be different for each report grade.
 
-* {cc_author = yes/no/fail/ask} -- should module authors should be sent a copy of 
-the test report at their {author@cpan.org} address (default: fail)
-* {edit_report = yes/no/fail/ask} -- edit the test report before sending 
-(default: ask/no)
-* {send_report = yes/no/fail/ask} -- should test reports be sent at all 
-(default: ask/yes)
-* {smtp_server = <server list>} -- one or more alternate outbound mail servers
-if the default perl.org mail servers cannot be reached (e.g. users behind a 
-firewall); multiple servers may be given, separated with a space 
-(default: none)
+Valid actions, and their associated meaning, are as follows:
+
+* {yes} -- automatic yes
+* {no} -- automatic no
+* {ask/no} or just {ask} -- ask each time, but default to no
+* {ask/yes} -- ask each time, but default to yes
+
+For "ask" prompts, the default will be used if return is pressed immediately at
+the prompt or if the {PERL_MM_USE_DEFAULT} environment variable is set to a
+true value.
+
+Action prompt options take one or more space-separated "grade:action" pairs,
+which are processed left to right.
+
+ edit_report = fail:ask/yes pass:no
+ 
+An action by itself is taken as a default to be used for any grade which does
+not have a grade-specific action.  A default action may also be set by using
+the word "default" in place of a grade.  
+
+ edit_report = ask/no
+ edit_report = default:ask/no
+ 
+A grade by itself is taken to have the action "yes" for that grade.
+
+ edit_report = default:no fail
+
+Multiple grades may be specified together by separating them with a slash.
+
+ edit_report = pass:no fail/na/unknown:ask/yes
+
+The action prompt options are:
+
+* {cc_author = <grade:action> ...} -- should module authors should be sent a copy of 
+the test report at their {author@cpan.org} address? (default:yes pass:no)
+* {edit_report = <grade:action> ...} -- edit the test report before sending? 
+(default:ask/no pass:no)
+* {send_report = <grade:action> ...} -- should test reports be sent at all?
+(default:ask/yes pass:yes)
+
+These options are included in the starter config file created automatically the
+first time CPAN::Reporter is configured interactively.
 
 Note that if {send_report} is set to "no", CPAN::Reporter will still go through
 the motions of preparing a report, but will discard it rather than send it.
-This is used for testing CPAN::Reporter.
+This is used primarily for testing CPAN::Reporter.
 
 A better way to disable CPAN::Reporter temporarily is with the CPAN option
 {test_report}:
 
  cpan> o conf test_report 0
- 
+
+== Mail Server
+
+By default, Test::Reporter attempts to send mail directly to perl.org mail 
+servers.  This may fail if a user's computer is behind a network firewall 
+that blocks outbound email.  In this case, the following option should
+be set to the outbound mail server (i.e., SMTP server) as provided by
+the user's Internet service provider (ISP):
+
+* {smtp_server = <server list>} -- one or more alternate outbound mail servers
+if the default perl.org mail servers cannot be reached; multiple servers may be
+given, separated with a space (none by default)
+
+In at least one reported case, an ISP's outbound mail servers also refused 
+to forward mail unless the {email_from} was from the ISP-given email address. 
+
+This option is also included in the starter config file.
+
 == Additional Options
 
 These additional options are only necessary in special cases, such as for
 testing, debugging or if a default editor cannot be found.
 
 * {email_to = <email address>} -- alternate destination for reports instead of
-{cpan-testers@perl.org}; used for testing (default: none)
+{cpan-testers@perl.org}; used for testing
 * {editor = <editor>} -- editor to use to edit the test report; if not set,
 Test::Reporter will use environment variables {VISUAL}, {EDITOR} or {EDIT}
-(in that order) to find an editor (default: none)
-* {debug = <boolean>} -- turns debugging on/off (default: off)
+(in that order) to find an editor 
+* {debug = <boolean>} -- turns debugging on/off
 
 = FUNCTIONS
 
